import logging

from django.apps import AppConfig
from core.utils import ConfigUtilMixin

MODULE_NAME = 'invoice'

DEFAULT_CONFIG = {
    "default_currency_code": "USD",
    "gql_invoice_search_perms": ["155101"],
    "gql_invoice_create_perms": ["155102"],
    "gql_invoice_update_perms": ["155103"],
    "gql_invoice_delete_perms": ["155104"],
    "gql_invoice_amend_perms": ["155109"],

    "gql_invoice_payment_search_perms": ["155201"],
    "gql_invoice_payment_create_perms": ["155202"],
    "gql_invoice_payment_update_perms": ["155203"],
    "gql_invoice_payment_delete_perms": ["155204"],
    "gql_invoice_payment_refund_perms": ["155206"],

    "gql_invoice_event_search_perms": ["155301"],
    "gql_invoice_event_create_perms": ["155302"],
    "gql_invoice_event_update_perms": ["155303"],
    "gql_invoice_event_delete_perms": ["155304"],
    "gql_invoice_event_create_message_perms": ["155306"],
    "gql_invoice_event_delete_my_message_perms": ["155307"],
    "gql_invoice_event_delete_all_message_perms": ["155308"],

    "gql_bill_search_perms": ["156101"],
    "gql_bill_create_perms": ["156102"],
    "gql_bill_update_perms": ["156103"],
    "gql_bill_delete_perms": ["156104"],
    "gql_bill_amend_perms": ["156109"],

    "gql_bill_payment_search_perms": ["156201"],
    "gql_bill_payment_create_perms": ["156202"],
    "gql_bill_payment_update_perms": ["156203"],
    "gql_bill_payment_delete_perms": ["156204"],
    "gql_bill_payment_refund_perms": ["156206"],

    "gql_bill_event_search_perms": ["156301"],
    "gql_bill_event_create_perms": ["156302"],
    "gql_bill_event_update_perms": ["156303"],
    "gql_bill_event_delete_perms": ["156304"],
    "gql_bill_event_create_message_perms": ["156306"],
    "gql_bill_event_delete_my_message_perms": ["156307"],
    "gql_bill_event_delete_all_message_perms": ["156308"],

    # Functions of type Callable[[QuerySet, User], QuerySet], to be used as custom user filters for bills and invoices
    # To be specified as "module_name.submodule.function_name"
    "bill_user_filter_function": None,
    "invoice_user_filter_function": None,
}

logger = logging.getLogger(__name__)


class InvoiceConfig(AppConfig, ConfigUtilMixin):
    name = MODULE_NAME

    default_currency_code = None
    gql_invoice_search_perms = None
    gql_invoice_create_perms = None
    gql_invoice_update_perms = None
    gql_invoice_delete_perms = None
    gql_invoice_amend_perms = None
    gql_invoice_payment_search_perms = None
    gql_invoice_payment_create_perms = None
    gql_invoice_payment_update_perms = None
    gql_invoice_payment_delete_perms = None
    gql_invoice_payment_refund_perms = None
    gql_invoice_event_search_perms = None
    gql_invoice_event_create_perms = None
    gql_invoice_event_update_perms = None
    gql_invoice_event_delete_perms = None
    gql_invoice_event_create_message_perms = None
    gql_invoice_event_delete_my_message_perms = None
    gql_invoice_event_delete_all_message_perms = None
    gql_bill_search_perms = None
    gql_bill_create_perms = None
    gql_bill_update_perms = None
    gql_bill_delete_perms = None
    gql_bill_amend_perms = None
    gql_bill_payment_search_perms = None
    gql_bill_payment_create_perms = None
    gql_bill_payment_update_perms = None
    gql_bill_payment_delete_perms = None
    gql_bill_payment_refund_perms = None
    gql_bill_event_search_perms = None
    gql_bill_event_create_perms = None
    gql_bill_event_update_perms = None
    gql_bill_event_delete_perms = None
    gql_bill_event_create_message_perms = None
    gql_bill_event_delete_my_message_perms = None
    gql_bill_event_delete_all_message_perms = None

    bill_user_filter = None
    invoice_user_filter = None

    def ready(self):
        from core.models import ModuleConfiguration
        cfg = ModuleConfiguration.get_or_default(MODULE_NAME, DEFAULT_CONFIG)
        self._load_config_fields(cfg)
        if cfg['bill_user_filter_function']:
            self._load_config_function('bill_user_filter', cfg['bill_user_filter_function'])
        if cfg['invoice_user_filter_function']:
            self._load_config_function('invoice_user_filter', cfg['invoice_user_filter_function'])



import logging

import graphene
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import ValidationError

from core.gql.gql_mutations.base_mutation import BaseMutation, BaseHistoryModelDeleteMutationMixin
from core.schema import OpenIMISMutation
from invoice.apps import InvoiceConfig
from invoice.models import Bill

logger = logging.getLogger(__name__)


class DeleteBillMutation(BaseHistoryModelDeleteMutationMixin, BaseMutation):
    _mutation_class = "DeleteBillMutation"
    _mutation_module = "invoice"
    _model = Bill

    @classmethod
    def _validate_mutation(cls, user, **data):
        if type(user) is AnonymousUser or not user.id or not user.has_perms(
                InvoiceConfig.gql_bill_delete_perms):
            raise ValidationError("mutation.authentication_required")

    class Input(OpenIMISMutation.Input):
        uuids = graphene.List(graphene.UUID)



import uuid

from core.models import MutationLog
from invoice.models import InvoiceEventMutation, InvoiceEvent
from invoice.tests.gql.base import InvoiceGQLTestCase


class InvoiceEventGQLTest(InvoiceGQLTestCase):

    search_for_message = '''
    query {{
  invoiceEvent(message_Istartswith: "{message_content}") {{
    edges {{
      node {{
        message
      }}
    }}
  }}
}}
'''

    create_mutation_str = '''  
mutation {{
  createInvoiceEventMessage(input:{{invoiceId: "{invoice_id}", message:"Some message with type", eventType: 0, clientMutationId: "{mutation_id}"}}) {{
    internalId, 
    clientMutationId
  }}
}}
'''

    def test_fetch_invoice_query(self):
        mutation_client_id = str(uuid.uuid4())
        mutation = self.create_mutation_str.format(invoice_id=self.invoice.id, mutation_id=mutation_client_id)
        self.graph_client.execute(mutation, context=self.BaseTestContext(self.user))

        output = self.graph_client.execute(
            self.search_for_message.format(message_content="Some message with"), context=self.BaseTestContext(self.user)
        )

        expected = \
            {'data': {
                'invoiceEvent': {
                    'edges': [
                        {'node': {
                            'message': 'Some message with type'
        }}]}}}
        self.assertEqual(output, expected)

    def test_create_payment_mutation(self):
        mutation_client_id = str(uuid.uuid4())
        mutation = self.create_mutation_str.format(invoice_id=self.invoice.id, mutation_id=mutation_client_id)
        self.graph_client.execute(mutation, context=self.BaseTestContext(self.user))
        mutation_log = MutationLog.objects.filter(client_mutation_id=mutation_client_id).first()
        obj: InvoiceEvent = InvoiceEventMutation.objects.get(mutation_id=mutation_log.id).invoice_event
        self.assertEqual(obj.invoice, self.invoice)
        self.assertEqual(obj.message, "Some message with type")
        self.assertEqual(obj.event_type, InvoiceEvent.EventType.MESSAGE)



# Generated by Django 3.0.14 on 2021-11-17 09:04

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('invoice', '0001_initial_migration'),
    ]

    operations = [
        migrations.AddField(
            model_name='billitem',
            name='deduction',
            field=models.DecimalField(db_column='Deduction', decimal_places=2, default=0.0, max_digits=18),
        ),
        migrations.AddField(
            model_name='historicalbillitem',
            name='deduction',
            field=models.DecimalField(db_column='Deduction', decimal_places=2, default=0.0, max_digits=18),
        ),
        migrations.AddField(
            model_name='historicalinvoicelineitem',
            name='deduction',
            field=models.DecimalField(db_column='Deduction', decimal_places=2, default=0.0, max_digits=18),
        ),
        migrations.AddField(
            model_name='invoicelineitem',
            name='deduction',
            field=models.DecimalField(db_column='Deduction', decimal_places=2, default=0.0, max_digits=18),
        ),
    ]



import graphene

from core.schema import signal_mutation_module_validate
from invoice.gql import query_mixins
from invoice.gql.invoice import DeleteInvoiceMutation, GenerateTimeframeInvoices
from invoice.gql.invoice_event.mutation import CreateInvoiceEventMutation
from invoice.gql.invoice_payment.mutation import (
    CreateInvoicePaymentMutation,
    UpdateInvoicePaymentMutation,
    DeleteInvoicePaymentMutation
)
from invoice.gql.bill.mutation import DeleteBillMutation
from invoice.gql.bill_event.mutation import CreateBillEventMutation
from invoice.gql.bill_payment.mutation import (
    CreateBillPaymentMutation,
    UpdateBillPaymentMutation,
    DeleteBillPaymentMutation
)
from invoice.gql.payment_invoice.mutation import (
    CreatePaymentInvoiceMutation,
    CreatePaymentInvoiceWithDetailMutation,
    UpdatePaymentInvoiceMutation,
    DeletePaymentInvoiceMutation
)
from invoice.gql.detail_payment_invoice.mutation import (
    CreateDetailPaymentInvoiceMutation,
    UpdateDetailPaymentInvoiceMutation,
    DeleteDetailPaymentInvoiceMutation
)
from invoice.models import (
    InvoicePayment,
    InvoicePaymentMutation,
    InvoiceEventMutation,
    InvoiceEvent,
    BillPayment,
    BillPaymentMutation,
    BillEventMutation,
    BillEvent,
    PaymentInvoice,
    DetailPaymentInvoice,
    PaymentInvoiceMutation,
    DetailPaymentInvoiceMutation,
)


class Query(
    query_mixins.InvoiceQueryMixin,
    query_mixins.InvoiceLineItemQueryMixin,
    query_mixins.InvoicePaymentQueryMixin,
    query_mixins.InvoiceEventQueryMixin,
    query_mixins.BillQueryMixin,
    query_mixins.BillItemQueryMixin,
    query_mixins.BillPaymentQueryMixin,
    query_mixins.BillEventQueryMixin,
    query_mixins.PaymentInvoiceQueryMixin,
    query_mixins.DetailPaymentInvoiceQueryMixin,
    graphene.ObjectType
):
    pass


class Mutation(graphene.ObjectType):
    # invoice mutations
    generate_invoices_for_time_period = GenerateTimeframeInvoices.Field()
    delete_invoice = DeleteInvoiceMutation.Field()
    create_invoice_payment = CreateInvoicePaymentMutation.Field()
    update_invoice_payment = UpdateInvoicePaymentMutation.Field()
    delete_invoice_payment = DeleteInvoicePaymentMutation.Field()

    create_invoice_event_message = CreateInvoiceEventMutation.Field()

    # bill mutations
    delete_bill = DeleteBillMutation.Field()
    create_bill_payment = CreateBillPaymentMutation.Field()
    update_bill_payment = UpdateBillPaymentMutation.Field()
    delete_bill_payment = DeleteBillPaymentMutation.Field()
    create_bill_event_type = CreateBillEventMutation.Field()

    # payment mutations
    create_payment_invoice = CreatePaymentInvoiceMutation.Field()
    create_payment_with_detail_invoice = CreatePaymentInvoiceWithDetailMutation.Field()
    update_payment_invoice = UpdatePaymentInvoiceMutation.Field()
    delete_payment_invoice = DeletePaymentInvoiceMutation.Field()
    create_detail_payment_invoice = CreateDetailPaymentInvoiceMutation.Field()
    update_detail_payment_invoice = UpdateDetailPaymentInvoiceMutation.Field()
    delete_detail_payment_invoice = DeleteDetailPaymentInvoiceMutation.Field()


def _on_mutation_log(mutation_model, model, obj_type, sender, **kwargs):
    uuids = kwargs['data'].get('ids', []) or kwargs['data'].get('uuids', [])
    if not uuids:
        uuid = kwargs['data'].get('id', None) or kwargs['data'].get('uuid', [])
        uuids = [uuid] if uuid else []
    if not uuids:
        return []
    impacted = model.objects.filter(uuid__in=uuids).all()
    for item in impacted:
        mutation_model.objects.create(
            **{obj_type: item, 'mutation_id':kwargs['mutation_log_id']}
        )
    return []


def on_invoice_payment_mutation(sender, **kwargs):
    if kwargs.get('mutation_class', None) \
            in ('CreateInvoicePaymentMutation', 'UpdateInvoicePaymentMutation', 'DeleteInvoicePaymentMutation'):
        return _on_mutation_log(InvoicePaymentMutation, InvoicePayment, 'invoice_payment', sender, **kwargs)

    if kwargs.get('mutation_class', None) in ('CreateInvoiceEventMutation'):
        return _on_mutation_log(InvoiceEventMutation, InvoiceEvent, 'invoice_event', sender, **kwargs)

    return []


def on_bill_payment_mutation(sender, **kwargs):
    if kwargs.get('mutation_class', None) \
            in ('CreateBillPaymentMutation', 'UpdateBillPaymentMutation', 'DeleteBillPaymentMutation'):
        return _on_mutation_log(BillPaymentMutation, BillPayment, 'bill_payment', sender, **kwargs)

    if kwargs.get('mutation_class', None) in ('CreateBillEventMutation'):
        return _on_mutation_log(BillEventMutation, BillEvent, 'bill_event', sender, **kwargs)

    return []


def on_payment_invoice_mutation(sender, **kwargs):
    if kwargs.get('mutation_class', None) \
            in ('CreatePaymentInvoiceMutation', 'UpdatePaymentInvoiceMutation', 'DeletePaymentInvoiceMutation',
                'CreatePaymentInvoiceWithDetailMutation'):
        return _on_mutation_log(PaymentInvoiceMutation, PaymentInvoice, 'payment_invoice', sender, **kwargs)

    if kwargs.get('mutation_class', None) \
            in ('CreateDetailPaymentInvoiceMutation', 'UpdateDetailPaymentInvoiceMutation',
                'DeleteDetailPaymentInvoiceMutation'):
        return _on_mutation_log(DetailPaymentInvoiceMutation, DetailPaymentInvoice, 'detail_payment_invoice', sender, **kwargs)

    return []


signal_mutation_module_validate["invoice"].connect(on_invoice_payment_mutation)
signal_mutation_module_validate["invoice"].connect(on_bill_payment_mutation)
signal_mutation_module_validate["invoice"].connect(on_payment_invoice_mutation)
